/*$Id: d_coil.cc,v 26.83 2008/06/05 04:46:59 al Exp $ -*- C++ -*-
 * Copyright (C) 2001 Albert Davis
 * Author: Albert Davis <aldavis@gnu.org>
 *
 * This file is part of "Gnucap", the Gnu Circuit Analysis Package
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *------------------------------------------------------------------
 * inductors
 * y.x = amps,  y.f0 = flux,  ev = y.f1 = henrys
 * q = y history in time
 * i.x = amps,  i.f0 = volts,      i.f1 = ohms
 * m.x = volts, m.c0 = amps, acg = m.c1 = mhos
 */
//testing=script 2006.07.17
#include "globals.h"
#include "e_ccsrc.h"
#include "e_storag.h"
/*--------------------------------------------------------------------------*/
namespace {
/*--------------------------------------------------------------------------*/
class DEV_INDUCTANCE : public STORAGE {
private:
  explicit DEV_INDUCTANCE(const DEV_INDUCTANCE& p) :STORAGE(p) {}
public:
  explicit DEV_INDUCTANCE()	:STORAGE() {}
public: // override virtual
  char	   id_letter()const	{return 'L';}
  std::string value_name()const {return "l";}
  std::string dev_type()const	{return "inductor";}
  int	   max_nodes()const	{return 2;}
  int	   min_nodes()const	{return 2;}
  int	   matrix_nodes()const	{return 2;}
  int	   net_nodes()const	{return 2;}
  bool	   use_obsolete_callback_parse()const {return true;}
  CARD*	   clone()const		{return new DEV_INDUCTANCE(*this);}
  void	   elaborate();
  //void   map_nodes();

  void	   tr_iwant_matrix();
  void	   tr_begin();
  void	   tr_restore();
  void     dc_advance();
  void     tr_advance();
  void     tr_regress();
  bool     tr_needs_eval()const;
  void     tr_queue_eval();
  bool	   do_tr();
  void	   tr_load();
  TIME_PAIR tr_review();
  //void   tr_accept();		//CARD/nothing
  void	   tr_unload();
  double   tr_involts()const	{return tr_outvolts();}
  double   tr_input()const	{return _m0.c0 + _m0.c1 * tr_involts();}
  double   tr_involts_limited()const {return tr_outvolts_limited();}
  double   tr_input_limited()const {return _m0.c0+_m0.c1*tr_involts_limited();}
  //double tr_amps()const	//ELEMENT
  double   tr_probe_num(const std::string&)const;

  void	   ac_iwant_matrix();
  void	   ac_begin();
  void	   do_ac();
  void	   ac_load();
  COMPLEX  ac_involts()const	{return ac_outvolts();}
  //COMPLEX ac_amps()const;	//ELEMENT
  //XPROBE ac_probe_ext(const std::string&)const;//ELEMENT

  std::string port_name(int i)const {
    assert(i >= 0);
    assert(i < 2);
    static std::string names[] = {"p", "n"};
    return names[i];
  }
};
/*--------------------------------------------------------------------------*/
class DEV_MUTUAL_L : public COMPONENT {
private:
  std::string		_output_label;
  DEV_INDUCTANCE*	_output;
  std::string		_input_label;
  DEV_INDUCTANCE*	_input;
private:
  explicit	DEV_MUTUAL_L(const DEV_MUTUAL_L& p);
public:
  explicit	DEV_MUTUAL_L();
private: // override virtual
  char	   id_letter()const	{return 'K';}
  bool	   print_type_in_spice()const {return false;}
  std::string value_name()const {return "k";}
  std::string dev_type()const	{itested(); return "mutual_inductor";}
  int	    max_nodes()const	{return 2;}
  int	    ext_nodes()const	{return 0;}
  int	    min_nodes()const	{return 2;}
  //int	   matrix_nodes()const	//CARD 0
  int	   net_nodes()const	{return 0;}
  int	   num_current_ports()const {return 2;}
  bool	   is_2port()const	{untested(); return true;}
  CARD*	   clone()const		{ return new DEV_MUTUAL_L(*this);}
  void	   elaborate();
  void	   map_nodes()		{}

  //void   tr_begin();		//CARD/nothing
  //void   tr_restore();	//CARD/nothing
  //void   dc_advance();	//CARD/nothing
  //void   tr_advance();	//CARD/nothing
  //void   tr_regress();	//CARD/nothing
  //bool   tr_needs_eval();	//CARD/nothing
  //void   tr_queue_eval();	//CARD/nothing
  //bool   do_tr();		//CARD/nothing
  //void   tr_load();		//CARD/nothing
  //TIME_PAIR tr_review();	//CARD/nothing
  //void   tr_accept();		//CARD/nothing
  //void   tr_unload();		//CARD/nothing
  //double tr_probe_num(const std::string&)const;//CKT_BASE/nothing

  //void   ac_begin();		//CARD/nothing
  //void   do_ac();		//CARD/nothing
  //void   ac_load();		//CARD/nothing
  COMPLEX  ac_involts()const {untested();return _n[IN1]->vac()-_n[IN2]->vac();}
  //XPROBE ac_probe_ext(const std::string&)const;//CKT_BASE/nothing

  void	   set_port_by_name(std::string& Name, std::string& Value)
		{untested(); COMPONENT::set_port_by_name(Name,Value);}
  void	   set_port_by_index(int Index, std::string& Value)
		{set_current_port_by_index(Index, Value);}
  bool	   node_is_connected(int i)const {
    switch (i) {
    case 0:  return _output_label != "";
    case 1:  return _input_label != "";
    default: unreachable(); return false;
    }
  }

  std::string port_name(int)const {
    return "";
  }
  std::string current_port_name(int i)const {
    assert(i >= 0);
    assert(i < 2);
    static std::string names[] = {"l1", "l2"};
    return names[i];
  }
  const std::string current_port_value(int i)const {
    switch (i) {
    case 0:  return _output_label;
    case 1:  return _input_label;
    default: unreachable(); return COMPONENT::current_port_value(i);
    }
  }
  void set_current_port_by_index(int i, const std::string& s) {
    switch (i) {
    case 0:  _output_label = s;	break;
    case 1:  _input_label = s;	break;
    default: unreachable();	break;
    }
  }
};
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
DEV_MUTUAL_L::DEV_MUTUAL_L()
  :COMPONENT(),
   _output_label(),
   _output(0),
   _input_label(),
   _input(0)
{
}
/*--------------------------------------------------------------------------*/
DEV_MUTUAL_L::DEV_MUTUAL_L(const DEV_MUTUAL_L& p)
  :COMPONENT(p),
   _output_label(p._output_label),
   _output(p._output),
   _input_label(p._input_label),
   _input(p._input)
{
}
/*--------------------------------------------------------------------------*/
// replace both primary and secondary (both simple L's)
// with the CCCS - L equivalent
// only works for 2.
void DEV_MUTUAL_L::elaborate()
{
  COMPONENT::elaborate();

  _output = dynamic_cast<DEV_INDUCTANCE*>(find_in_my_scope(_output_label));
  if (!_output) {
    throw Exception_Type_Mismatch(long_label(), _output_label, "inductor");
  }else if (_output->has_common()) {
    throw Exception(long_label() + ": " + _output_label + " must be linear");
  }else{
  }
  _input = dynamic_cast<DEV_INDUCTANCE*>(find_in_my_scope(_input_label));
  if (!_input) {
    throw Exception_Type_Mismatch(long_label(), _input_label, "inductor");
  }else if (_input->has_common()) {
    throw Exception(long_label() + ": " + _input_label + " must be linear");
  }else{
  }
  _output->new_subckt();
  _input->new_subckt();

  double l1 = _output->value();
  double l2 = _input->value();
  double lm  = value() * sqrt(l1 * l2);
  double det = l1 * l2 - lm * lm;
  {
    DEV_INDUCTANCE* pri = prechecked_cast<DEV_INDUCTANCE*>(_output->clone());
    assert(pri);
    pri->set_owner(_output);
    pri->set_value(det / l2);
    _output->subckt()->push_back(pri);

    const CARD* proto = device_dispatcher["cccs"];
    assert(proto);
    CCSRC_BASE* sub = dynamic_cast<CCSRC_BASE*>(proto->clone());
    assert(sub);
    sub->set_parameters_cc("F" + pri->short_label(), _input, NULL,
			   -lm / l2, _input->_n[OUT1], _input->_n[OUT2], pri);
    _input->subckt()->push_back(sub);
  }
  {
    DEV_INDUCTANCE* sec = prechecked_cast<DEV_INDUCTANCE*>(_input->clone());
    assert(sec);
    sec->set_owner(_input);
    sec->set_value(det / l1);
    _input->subckt()->push_back(sec).elaborate();
    
    const CARD* proto = device_dispatcher["cccs"];
    assert(proto);
    CCSRC_BASE* sub = dynamic_cast<CCSRC_BASE*>(proto->clone());
    assert(sub);
    sub->set_parameters_cc("F" + sec->short_label(), _output, NULL,
			   -lm / l1, _output->_n[OUT1], _output->_n[OUT2], sec);
    _output->subckt()->push_back(sub);
  }
  _input->subckt()->elaborate();
  _output->subckt()->elaborate();
  
  assert(!is_constant()); /* because of integration */
}
/*--------------------------------------------------------------------------*/
void DEV_INDUCTANCE::tr_iwant_matrix()
{
  if (subckt()) {
    subckt()->tr_iwant_matrix();
  }else{
    tr_iwant_matrix_passive();
  }
}
/*--------------------------------------------------------------------------*/
void DEV_INDUCTANCE::elaborate()
{
  if (subckt()) {
    //subckt()->elaborate(); // redundant
  }else{
    STORAGE::elaborate();
  }
}
/*--------------------------------------------------------------------------*/
void DEV_INDUCTANCE::tr_begin()
{
  if (subckt()) {
    subckt()->tr_begin();
  }else{
    STORAGE::tr_begin();
  }
}
/*--------------------------------------------------------------------------*/
void DEV_INDUCTANCE::tr_restore()
{
  if (subckt()) {untested();
    subckt()->tr_restore();
  }else{
    STORAGE::tr_restore();
  }
}
/*--------------------------------------------------------------------------*/
void DEV_INDUCTANCE::dc_advance()
{
  if (subckt()) {
    subckt()->dc_advance();
  }else{
    STORAGE::dc_advance();
  }
}
/*--------------------------------------------------------------------------*/
void DEV_INDUCTANCE::tr_advance()
{
  if (subckt()) {
    subckt()->tr_advance();
  }else{
    STORAGE::tr_advance();
  }
}
/*--------------------------------------------------------------------------*/
void DEV_INDUCTANCE::tr_regress()
{
  if (subckt()) {
    subckt()->tr_regress();
  }else{
    STORAGE::tr_regress();
  }
}
/*--------------------------------------------------------------------------*/
bool DEV_INDUCTANCE::tr_needs_eval()const
{
  assert(!is_q_for_eval());
  if (subckt()) {untested();
    return subckt()->tr_needs_eval();
  }else{
    return true;
    //BUG// errors when it has a current probe, as in mutual inductance
    //work around: always return true.
    //return STORAGE::tr_needs_eval();
  }
}
/*--------------------------------------------------------------------------*/
void DEV_INDUCTANCE::tr_queue_eval()
{
  if (subckt()) {
    if (subckt()->tr_needs_eval()) {
      subckt()->tr_queue_eval();
    }else{untested();
    }
  }else{
    STORAGE::tr_queue_eval();
  }
}
/*--------------------------------------------------------------------------*/
bool DEV_INDUCTANCE::do_tr()
{
  if (subckt()) {untested();
    set_converged(subckt()->do_tr());
  }else{
    double x = NOT_VALID;
    if (using_tr_eval()) {
      //x = tr_involts_limited();
      _y[0].x = tr_input_limited(); // _m0.c0 + _m0.c1 * x;
      tr_eval();
      if (_y[0].f1 == 0.) {untested();
	error(bDANGER, long_label() + ": short circuit,  L = 0\n");
	_y[0].f1 = OPT::shortckt;
	set_converged(conv_check());
      }else{
      }
    }else{
      //x = _n[OUT1].v0() - _n[OUT2].v0();
      //x = tr_involts();
      _y[0].x = tr_input(); // _m0.c0 + _m0.c1 * x;
      assert(_y[0].f1 == value());
      _y[0].f0 = _y[0].x * _y[0].f1;
      assert(converged());
    }
    store_values();
    q_load();

    // i is really voltage ..
    // _i[0].x = current, _i[0].f0 = voltage, _i[0].f1 = ohms
    _i[0] = differentiate();
    _m0.x = x;
    _m0.c1 = 1 / ((_i[0].c1()==0) ? OPT::shortckt : _i[0].c1());
    _m0.c0 = -_i[0].c0() * _m0.c1;
  }
  return converged();
}
/*--------------------------------------------------------------------------*/
void DEV_INDUCTANCE::tr_load()
{
  if (subckt()) {
    subckt()->tr_load();
  }else{
    tr_load_passive();
  }
}
/*--------------------------------------------------------------------------*/
TIME_PAIR DEV_INDUCTANCE::tr_review()
{
  if (subckt()) {
    return _time_by = subckt()->tr_review();
  }else{
    return STORAGE::tr_review();
  }
}
/*--------------------------------------------------------------------------*/
void DEV_INDUCTANCE::tr_unload()
{untested();
  if (subckt()) {untested();
    subckt()->tr_unload();
  }else{untested();
    tr_unload_passive();
  }
}
/*--------------------------------------------------------------------------*/
void DEV_INDUCTANCE::ac_iwant_matrix()
{
  if (subckt()) {
    subckt()->ac_iwant_matrix();
  }else{
    ac_iwant_matrix_passive();
  }
}
/*--------------------------------------------------------------------------*/
void DEV_INDUCTANCE::ac_begin()
{
  if (subckt()) {
    subckt()->ac_begin();
  }else{
    _ev = _y[0].f1;
  }
}
/*--------------------------------------------------------------------------*/
void DEV_INDUCTANCE::do_ac()
{
  if (subckt()) {
    subckt()->do_ac();
  }else{
    if (using_ac_eval()) {
      ac_eval();
    }else{
      assert(_ev == _y[0].f1);
      assert(has_tr_eval() || _ev == double(value()));
    }
    if (_ev * SIM::jomega == 0.) {itested();
      _acg = 1. / OPT::shortckt;
    }else{
      _acg = 1. / (_ev * SIM::jomega);
    }
    ac_load();
  }
}
/*--------------------------------------------------------------------------*/
void DEV_INDUCTANCE::ac_load()
{
  if (subckt()) {
    unreachable();
    subckt()->ac_load();
  }else{
    ac_load_passive();
  }
}
/*--------------------------------------------------------------------------*/
double DEV_INDUCTANCE::tr_probe_num(const std::string& x)const
{
  if (subckt()) {
    return COMPONENT::tr_probe_num(x);
  }else{
    if (Umatch(x, "flux ")) {untested();
      return _y[0].f0;
    }else if (Umatch(x, "ind{uctance} |l ")) {untested();
      return _y[0].f1;
    }else if (Umatch(x, "dldt ")) {untested();
      return (_y[0].f1 - _y[1].f1) / _dt;
    }else if (Umatch(x, "dl ")) {untested();
      return (_y[0].f1 - _y[1].f1);
    }else if (Umatch(x, "dfdt ")) {untested();
      return (_y[0].f0 - _y[1].f0) / _dt;
    }else if (Umatch(x, "dflux ")) {untested();
      return (_y[0].f0 - _y[1].f0);
    }else{
      return STORAGE::tr_probe_num(x);
    }
  }
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
DEV_MUTUAL_L   p1;
DEV_INDUCTANCE p2;
DISPATCHER<CARD>::INSTALL
  d1(&device_dispatcher, "K|mutual_inductor", &p1),
  d2(&device_dispatcher, "L|inductor",        &p2);
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
