/*$Id: m_expression.h,v 26.81 2008/05/27 05:34:00 al Exp $ -*- C++ -*-
 * Copyright (C) 2003 Albert Davis
 * Author: Albert Davis <aldavis@gnu.org>
 *
 * This file is part of "Gnucap", the Gnu Circuit Analysis Package
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *------------------------------------------------------------------
 */
#include "m_base.h"
/*--------------------------------------------------------------------------*/
class Symbol_Table;
class PARAM_LIST;
class Expression;
/*--------------------------------------------------------------------------*/
class Token
  :public Base
{
public:
  enum Type {SYMBOL, BINOP, STOP, PARLIST, UNARY, CONSTANT};
private:
  std::string _name;
  const Base* _data;
  std::string _args;
public:
  void parse(CS&) {unreachable();}
private:
  void dump(std::ostream&)const;
protected:
  explicit Token(const std::string Name, Type, const Base* Data,
		 const std::string Args)
    : _name(Name), _data(Data), _args(Args) {}
  explicit Token(const Token& P)
    : Base(), _name(P._name), _data(P._data), _args(P._args) {assert(!_data);}
public:
  virtual ~Token() {if (_data) {delete _data;}else{}}
  virtual Token*     clone()const = 0;
  const std::string& name()const {return _name;}
  //virtual Type	     type()const = 0;
  const Base*	     data()const {return _data;}
  const std::string& args()const {return _args;}
  const std::string  full_name()const {return name() + args();}
  virtual void	     stack_op(Expression*)const {unreachable();}
  bool operator==(const Token& P) {return (typeid(*this)==typeid(P))
      && (data()==P.data()) && (name()==P.name()) && (args()==P.args());}
};
/*--------------------------------------------------------------------------*/
class Token_SYMBOL : public Token
{
public:
  explicit Token_SYMBOL(const std::string Name, const std::string Args)
    : Token(Name, SYMBOL, NULL, Args) {}    
  explicit Token_SYMBOL(const Token_SYMBOL& P) : Token(P) {}
  Token* clone()const {return new Token_SYMBOL(*this);}
  Type type()const {return SYMBOL;}  
  void stack_op(Expression*)const;
};
class Token_BINOP : public Token
{
public:
  explicit Token_BINOP(const std::string Name)
    : Token(Name, BINOP, NULL, "") {}    
  explicit Token_BINOP(const Token_BINOP& P) : Token(P) {}
  Token* clone()const {return new Token_BINOP(*this);}
  Type   type()const {return BINOP;}  
  Token* op(const Token* t1, const Token* t2)const;
  void stack_op(Expression*)const;
};
class Token_STOP : public Token
{
public:
  explicit Token_STOP(const std::string Name)
    : Token(Name, STOP, NULL, "") {}    
  explicit Token_STOP(const Token_STOP& P) : Token(P) {}
  Token* clone()const {return new Token_STOP(*this);}
  Type type()const {return STOP;}  
  void stack_op(Expression*)const;
};
class Token_PARLIST : public Token
{
public:
  explicit Token_PARLIST(const std::string Name)
    : Token(Name, PARLIST, NULL, "") {}    
  explicit Token_PARLIST(const Token_PARLIST& P) : Token(P) {}
  Token* clone()const {return new Token_PARLIST(*this);}
  Type type()const {return PARLIST;}  
  void stack_op(Expression*)const;
};
class Token_UNARY : public Token
{
public:
  explicit Token_UNARY(const std::string Name)
    : Token(Name, UNARY, NULL, "") {}    
  explicit Token_UNARY(const Token_UNARY& P) : Token(P) {}
  Token* clone()const {return new Token_UNARY(*this);}
  Type   type()const {return UNARY;}  
  Token* op(const Token* t1)const;
  void stack_op(Expression*)const;
};
class Token_CONSTANT : public Token
{
public:
  explicit Token_CONSTANT(const std::string Name, const Base* Data, const std::string Args)
    : Token(Name, CONSTANT, Data, Args) {}    
  explicit Token_CONSTANT(const Token_CONSTANT& P) : Token(P) {}
  Token* clone()const {return new Token_CONSTANT(*this);}
  Type type()const {return CONSTANT;}  
  void stack_op(Expression*)const;
};
/*--------------------------------------------------------------------------*/
class Expression
  :public List_Base<Token>
{
public:
  PARAM_LIST* _params;
public:
  void parse(CS&);
  void dump(std::ostream&)const;
private: // expression-in.cc
  void arglisttail(CS& File);
  void arglist(CS& File);
  void leaf(CS& File);
  void factor(CS& File);
  void termtail(CS& File);
  void term(CS& File);
  void addexptail(CS& File);
  void addexp(CS& File);
  void logicaltail(CS& File);
  void logical(CS& File);
  void andtail(CS& File);
  void andarg(CS& File);
  void exptail(CS& File);
  void expression(CS& File);
public:
  explicit Expression() : _params(NULL) {}
  explicit Expression(CS& File) : _params(NULL) {parse(File);}
private: // expression-reduce.cc
  void reduce_copy(const Expression&);
public:
  explicit Expression(const Expression&, PARAM_LIST*);
public: // other
  bool as_bool()const {return (!is_empty() && back()->data());}
  double eval()const {
    const Float* f = dynamic_cast<const Float*>(back()->data());
    return ((f && size()==1) ? (f->value()) : (NOT_INPUT));
  }
};
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
