/*$Id: s_fo.cc,v 26.84 2008/06/08 21:27:02 al Exp $ -*- C++ -*-
 * Copyright (C) 2001 Albert Davis
 * Author: Albert Davis <aldavis@gnu.org>
 *
 * This file is part of "Gnucap", the Gnu Circuit Analysis Package
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *------------------------------------------------------------------
 * tran and fourier commands -- top
 * performs transient analysis, silently, then fft.
 * outputs results of fft
 */
//testing=script 2007.11.21
#include "m_phase.h"
#include "globals.h"
#include "declare.h"	/* plclose, plclear, fft */
#include "u_prblst.h"
#include "s_tr.h"
/*--------------------------------------------------------------------------*/
namespace {
/*--------------------------------------------------------------------------*/
class FOURIER : public TRANSIENT {
public:
  void	do_it(CS&, CARD_LIST*);
  explicit FOURIER():
    TRANSIENT(),
    _fstart(0.),
    _fstop(0.),
    _fstep(0.),
    _timesteps(0),
    _fdata(NULL)
  {
  }
  ~FOURIER() {}
private:
  explicit FOURIER(const FOURIER&): TRANSIENT() {unreachable(); incomplete();}
  void	setup(CS&);	/* s_fo_set.cc */
  void	fftallocate();
  void	fftunallocate();
  void	foout();	/* s_fo_out.cc */
  void	fohead(const PROBE&);
  void	foprint(COMPLEX*);
  void	store_results(double); // override virtual
private:
  double _fstart;	/* user start frequency */
  double _fstop;	/* user stop frequency */
  double _fstep;	/* fft frequecncy step */
  int    _timesteps;	/* number of time steps in tran analysis, incl 0 */
  COMPLEX** _fdata;	/* storage to allow postprocessing */
};
/*--------------------------------------------------------------------------*/
static	int	to_pow_of_2(double);
static  int	stepnum(double,double,double);
static	COMPLEX	find_max(COMPLEX*,int,int);
static	double  db(COMPLEX);
/*--------------------------------------------------------------------------*/
void FOURIER::do_it(CS& cmd, CARD_LIST*)
{
  set_command_fourier();
  reset_timers();
  ::status.four.reset().start();
  
  try {
    init();
    alloc_vectors();
    
    aa.reallocate();
    aa.dezero(OPT::gmin);
    aa.set_min_pivot(OPT::pivtol);
    
    lu.reallocate();
    lu.dezero(OPT::gmin);
    lu.set_min_pivot(OPT::pivtol);
    
    setup(cmd);
    fftallocate();
    
    ::status.set_up.stop();
    switch (ENV::run_mode) {untested();
    case rPRE_MAIN:	unreachable();		break;
    case rBATCH:	untested();
    case rINTERACTIVE:  itested();
    case rSCRIPT:	sweep(); foout();	break;
    case rPRESET:	untested(); /*nothing*/ break;
    }
  }catch (Exception& e) {itested();
    error(bDANGER, e.message() + '\n');
  }
  
  fftunallocate();
  unalloc_vectors();
  lu.unallocate();
  aa.unallocate();
  
  ::status.four.stop();
  ::status.total.stop();
  
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
/* store: stash time domain data in preparation for Fourier Transform
 */
void FOURIER::store_results(double x)
{
  TRANSIENT::store_results(x);

  if (step_cause() == scUSER) {
    int ii = 0;
    for (PROBELIST::const_iterator
	   p=printlist().begin();  p!=printlist().end();  ++p) {
      assert(stepno < _timesteps);
      _fdata[ii][stepno] = p->value();
      ++ii;
    }
  }else{untested();
  }
}
/*--------------------------------------------------------------------------*/
/* foout:  print out the results of the transform
 */
void FOURIER::foout()
{
  plclose();
  plclear();
  int ii = 0;
  for (PROBELIST::const_iterator
	 p=printlist().begin();  p!=printlist().end();  ++p) {
    fohead(*p);
    fft(_fdata[ii], _timesteps-1,  0);
    foprint(_fdata[ii]);
    ++ii;
  }
}
/*--------------------------------------------------------------------------*/
/* fo_head: print output header
 * arg is index into probe array, to select probe name
 */
void FOURIER::fohead(const PROBE& prob)
{
  out.form("# %-10s", prob.label().c_str())
    << "--------- actual ---------  -------- relative --------\n"
    << "#freq       "
    << "value        dB      phase  value        dB      phase\n";
}
/*--------------------------------------------------------------------------*/
/* fo_print: print results of fourier analysis
 * for all points at single probe
 */
void FOURIER::foprint(COMPLEX *data)
{
  int startstep = stepnum(0., _fstep, _fstart);
  assert(startstep >= 0);
  int stopstep  = stepnum(0., _fstep, _fstop );
  assert(stopstep < _timesteps);
  COMPLEX maxvalue = find_max(data,std::max(1,startstep),stopstep);
  if (maxvalue == 0.) {untested();
    maxvalue = 1.;
  }else{
  }
  data[0] /= 2;
  for (int ii = startstep;  ii <= stopstep;  ++ii) {
    double frequency = _fstep * ii;
    assert(ii >= 0);
    assert(ii < _timesteps);
    COMPLEX unscaled = data[ii];
    COMPLEX scaled = unscaled / maxvalue;
    unscaled *= 2;
    out.form("%s%s%7.2f %8.3f %s%7.2f %8.3f\n",
	     ftos(frequency,    11,5,out.format()),
        ftos(std::abs(unscaled),11,5,out.format()),
	     db(unscaled),
	     phase(unscaled*COMPLEX(0.,1)),
        ftos(std::abs(scaled),  11,5,out.format()),
	     db(scaled),
	     phase(scaled) ) ;
  }
}
/*--------------------------------------------------------------------------*/
/* stepnum: return step number given its frequency or time
 */
static int stepnum(double start, double step, double here)
{
  return int((here-start)/step + .5);
}
/*--------------------------------------------------------------------------*/
/* find_max: find the max magnitude in a COMPLEX array
 */
static COMPLEX find_max(COMPLEX *data, int start, int stop)
{
  COMPLEX maxvalue = 0.;
  for (int ii = start;  ii <= stop;  ++ii) {
    if (std::abs(data[ii]) > std::abs(maxvalue)) {
      maxvalue = data[ii];
    }else{
    }
  }
  return maxvalue;
}
/*--------------------------------------------------------------------------*/
static double db(COMPLEX value)
{
  return  20. * log10(std::max(std::abs(value),VOLTMIN));
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
/* fo_setup: fourier analysis: parse command string and set options
 * 	(options set by call to TRANSIENT::options)
 */
void FOURIER::setup(CS& cmd)
{
  cont = true;
  if (cmd.is_pfloat()) {
    double arg1 = cmd.ctopf();
    double arg2 = cmd.ctopf();
    double arg3 = cmd.ctopf();
    if (arg3 != 0.) {			    /* 3 args: all */
      _fstart = arg1;
      _fstop  = arg2;
      _fstep  = arg3;
    }else if (arg2 != 0.) {untested();	    /* 2 args: start = 0 */
      if (arg1 >= arg2) {untested();	    /* 2 args: stop, step */
	_fstart = 0.;		    	    /* 	(stop > step) */
	_fstop  = arg1;
	_fstep  = arg2;
      }else{untested(); /* arg1 < arg2 */    /* 2 args: step, stop */
	_fstart = 0.;
	_fstop  = arg2;
	_fstep  = arg1;
      }
    }else if (arg1 == 0.) {untested();	    /* 1 arg: start */
      _fstart = 0.;
      /* _fstop unchanged */
      /* _fstep unchanged */
    }else{untested();			    /* 1 arg: step */
      _fstart = 0.;
      _fstop  = OPT::harmonics * arg1;
      _fstep  = arg1;
    }
  }else{itested();
  }
  /* else (no args) : no change */
  
  if (_fstep == 0.) {itested();
    throw Exception("frequency step = 0");
  }else{
  }
  if (_fstop == 0.) {untested();
    _fstop = OPT::harmonics * _fstep;
  }else{
  }
  options(cmd);
  
  _timesteps = to_pow_of_2(_fstop*2 / _fstep) + 1;
  if (cold) {untested();
    cont = false;
    tstart = 0.;
  }else{
    cont = true;
    tstart = last_time;
  }
  tstop = tstart + 1. / _fstep;
  tstep = 1. / _fstep / (_timesteps-1);
  time1 = time0 = tstart;
  dtmax = tstep / static_cast<double>(skip);
  dtmin = std::max(dtmin,dtmax/dtratio);
  freq = _fstep;
}
/*--------------------------------------------------------------------------*/
/* allocate:  allocate space for fft
 */
void FOURIER::fftallocate()
{
  int probes = printlist().size();
  _fdata = new COMPLEX*[probes];
  for (int ii = 0;  ii < probes;  ++ii) {
    _fdata[ii] = new COMPLEX[_timesteps+100];
  }
}
/*--------------------------------------------------------------------------*/
/* unallocate:  unallocate space for fft
 */
void FOURIER::fftunallocate()
{
  if (_fdata) {
    for (int ii = 0;  ii < printlist().size();  ++ii) {
      delete [] _fdata[ii];
    }
    delete [] _fdata;
    _fdata = NULL;
  }else{itested();
  }
}
/*--------------------------------------------------------------------------*/
/* to_pow_of_2: round up to nearest power of 2
 * example: z=92 returns 128
 */
static int to_pow_of_2(double z)
{
  int x = static_cast<int>(floor(z));
  int y;
  for (y = 1; x > 0; x >>= 1) {
    y <<= 1;
  }
  return y;
}   
/*--------------------------------------------------------------------------*/
static FOURIER p3;
DISPATCHER<CMD>::INSTALL d3(&command_dispatcher, "fourier", &p3);
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
